@@ -1,11 +1,27 @@
 package Log::ger::Output::Composite;
 
-our $DATE = '2017-06-23'; # DATE
-our $VERSION = '0.002'; # VERSION
+our $DATE = '2017-06-25'; # DATE
+our $VERSION = '0.003'; # VERSION
 
 use strict;
 use warnings;
 
+sub _get_min_max_level {
+    my $level = shift;
+    my ($min, $max);
+    if (defined $level) {
+        if (ref $level eq 'ARRAY') {
+            $min = Log::ger::Util::numeric_level($level->[0]);
+            $max = Log::ger::Util::numeric_level($level->[1]);
+            ($min, $max) = ($max, $min) if $min > $max;
+        } else {
+            $min = 0;
+            $max = Log::ger::Util::numeric_level($level);
+        }
+    }
+    ($min, $max);
+}
+
 sub get_hooks {
     my %conf = @_;
 
@@ -41,7 +57,7 @@ sub get_hooks {
             # levels. so we handle all levels.
             __PACKAGE__, 5,
             sub {
-                no strict 'refs';
+                require Data::Dmp;
 
                 my %args = @_;
 
@@ -62,7 +78,7 @@ sub get_hooks {
                             $target, $target_arg, 'create_log_routine');
                         $saved_pt ||= $saved0;
                     }
-                    my $oconf = $ospec->{args} || {};
+                    my $oconf = $ospec->{conf} || {};
                     Log::ger::Util::set_plugin(
                         name => $oname,
                         prefix => 'Log::ger::Output::',
@@ -73,8 +89,8 @@ sub get_hooks {
                     my $res = Log::ger::run_hooks(
                         'create_log_routine', \%args, 1, $target, $target_arg,
                     );
-                    my $logger = $res or die "Hook from output module '$oname' ".
-                        "didn't produce log routine";
+                    my $logger = $res or die "Hook from output module ".
+                        "'$oname' didn't produce log routine";
                     push @$loggers, $logger;
                 }
                 Log::ger::Util::restore_hooks('create_log_routine', $saved_g)
@@ -91,7 +107,7 @@ sub get_hooks {
                 # string-eval'ed code
                 my ($addr) = "$loggers" =~ /\(0x(\w+)/;
                 my $varname = "Log::ger::Stash::$addr";
-                ${$varname} = $loggers;
+                { no strict 'refs'; ${$varname} = $loggers; }
 
                 # generate logger routine
                 my $logger;
@@ -99,30 +115,60 @@ sub get_hooks {
                     my @src;
                     push @src, "sub {\n";
 
-                    #push @src, "  my $ctx = $_[0];\n";
-
-                    # XXX filter by category_level
-
                     for my $i (0..$#ospecs) {
                         my $ospec = $ospecs[$i];
                         push @src, "  # output #$i: $ospec->{_name}\n";
                         push @src, "  {\n";
-                        # XXX filter by output's category_level
 
-                        # filter by output level
-                        if (defined $ospec->{level}) {
-                            push @src, "    last unless ".
-                                Log::ger::Util::numeric_level($ospec->{level}).
-                                  " >= $args{level};\n";
-                        } else {
-                            # filter by general level
-                            push @src, "  last if \$Log::ger::Current_Level < $args{level};\n";
+                        # filter by output's category_level and category-level
+                        if ($ospec->{category_level} || $conf{category_level}) {
+                            push @src, "    my \$cat = \$_[0]{category} || ".
+                                "'';\n";
+
+                            my @cats;
+                            if ($ospec->{category_level}) {
+                                for my $cat (keys %{$ospec->{category_level}}) {
+                                    my $clevel = $ospec->{category_level}{$cat};
+                                    push @cats, [$cat, 1, $clevel];
+                                }
+                            }
+                            if ($conf{category_level}) {
+                                for my $cat (keys %{$conf{category_level}}) {
+                                    my $clevel = $conf{category_level}{$cat};
+                                    push @cats, [$cat, 2, $clevel];
+                                }
+                            }
+
+                            for my $cat (sort {
+                                length($b->[0]) <=> length($a->[0]) ||
+                                    $a->[0] cmp $b->[0] ||
+                                        $a->[1] <=> $b->[1]} @cats) {
+                                push @src, "    if (\$cat eq ".Data::Dmp::dmp($cat->[0])." || index(\$cat, ".Data::Dmp::dmp("$cat->[0]\::").") == 0) { ";
+                                my ($min_level, $max_level) =
+                                    _get_min_max_level($cat->[2]);
+                                push @src, "if ($args{level} >= $min_level && ".
+                                    "$args{level} <= $max_level) { goto L } else { last }";
+                                push @src, " }\n";
+                            }
+                            push @src, "\n";
                         }
 
+                        # filter by output level
+                        my ($min_level, $max_level) = _get_min_max_level(
+                            $ospec->{level});
+                        if (defined $min_level) {
+                            push @src, "    if ($args{level} >= $min_level && ".
+                                "$args{level} <= $max_level) { goto L } else { last }\n";
+                        }
+
+                        # filter by general level
+                        push @src, "    if (\$Log::ger::Current_Level >= $args{level}) { goto L } else { last }\n";
+
                         # run output's log routine
-                        push @src, "    \$$varname\->[$i]->(\@_);\n";
-                        push @src, "  } # output #$i\n\n";
-                    }
+                        push @src, "    L: \$$varname\->[$i]->(\@_);\n";
+                        push @src, "  }\n";
+                        push @src, "  # end output #$i\n\n";
+                    } # for ospec
 
                     push @src, "};\n";
                     my $src = join("", @src);
@@ -150,7 +196,7 @@ Log::ger::Output::Composite - Composite output
 
 =head1 VERSION
 
-version 0.002
+version 0.003
 
 =head1 SYNOPSIS
 
@@ -159,22 +205,22 @@ version 0.002
          # single screen output
          Screen => {
              level => 'info', # set mper-output level. optional.
-             args => { use_color=>1 },
+             conf => { use_color=>1 },
          },
          # multiple file outputs
          File   => [
              {
+                 conf => { path=>'/var/log/myapp.log' },
                  level => 'warn',
                  # set per-category, per-output level. optional.
                  category_level => {
                      # don't log myapp.security messages to this file
                      'myapp.security' => 'off',
                  },
-                 args => { path=>'/var/log/myapp.log' },
              },
              {
-                 path => '/var/log/myapp-security.log',
-                 level => 'off',
+                 conf => { path => '/var/log/myapp-security.log' },
+                 level => 'warn',
                  category_level => {
                      # only myapp.security messages go to this file
                      'myapp.security' => 'warn',
@@ -195,11 +241,9 @@ version 0.002
 
 =head1 DESCRIPTION
 
-B<EARLY RELEASE>.
-
-This is a L<Log::ger> output that can multiplex output to multiple outputs and
-do filtering using per-category level, per-output level, or per-output
-per-category level.
+This is a L<Log::ger> output that can multiplex output to several outputs and do
+filtering per-category level, per-output level, or per-output per-category
+level.
 
 =for Pod::Coverage ^(.+)$
 
@@ -207,6 +251,40 @@ per-category level.
 
 =head2 outputs => hash
 
+Specify outputs. It's a hash with output name as keys and output specification
+as values.
+
+Output name is the name of output module without the C<Log::ger::Output::>
+prefix, e.g. L<Screen|Log::ger::Output::Screen> or
+L<File||Log::ger::Output::File>.
+
+Output specification is either a hashref or arrayref of hashrefs to specify
+multiple outputs per type (e.g. if you want to output to two File's). Known
+hashref keys:
+
+=over
+
+=item * conf => hashref
+
+Specify output configuration. See each output documentation for the list of
+available configuration parameters.
+
+=item * level => str|int|[min, max]
+
+Specify per-output level. If specified, logging will be done at this level
+instead of the general level. For example, if this is set to C<debug> then debug
+messages and higher will be sent to output even though the general level is
+C<warn>. Vice versa, if this is set to C<error> then even though the general
+level is C<warn>, warning messages won't be sent to this output; only C<error>
+messages and higher will be sent.
+
+You can specify a single level (e.g. 1 or "trace") or a two-element array to
+specify minimum and maximum level (e.g. C<<["trace", "info"]>>). If you
+accidentally mix up minimum and maximum, this module will helpfully fix it for
+you.
+
+=back
+
 =head2 category_level => hash
 
 =head2
